library ac_dart.host_service.http_client;

import 'dart:async';
import 'dart:io';

import 'package:atlassian_connect_host/host_event_model.dart';
import 'package:atlassian_connect_host/host_registration_persistence.dart';
import 'package:atlassian_connect_host/src/http_method.dart';
import 'package:atlassian_connect_host/src/preconditions.dart';
import 'package:atlassian_connect_jwt/atlassian_connect_jwt.dart';
import 'package:logging/logging.dart';
import 'package:option/option.dart';
import 'package:path/path.dart' as _path;

_path.Context get _url => _path.url;

final Logger _log = new Logger('AcHostHttpClient');

/**
 * An http client for communicating with Atlassian product hosts. Supports
 * signing requests with the required Jwt mechanism.
 * 
 * Note this is a wrapper for the core [HttpClient] class
 */
class AcHostHttpClient {
  final Option<String> _sharedSecret;
  final Option<String> _issuer;
  final Option<String> _audience;
  final Uri baseUri;
  final HttpClient _httpClient;
  final bool _useAuthHeader;
  final bool _sendAddonAsSubject;

  AcHostHttpClient.fromHostRego(
      AcHostRegistration hostRego, AddonHost addonHost,
      {bool sendAddonAsSubject: false})
      : this(
            checkNotNull(hostRego).hostDetails.baseUrl,
            new Some(hostRego.sharedSecret),
            new Some(addonHost.key),
            null,
            new Some(hostRego.hostDetails.key),
            true,
            sendAddonAsSubject ?? false);

  /// [httpClient] exposed for testing
  AcHostHttpClient(this.baseUri,
      [this._sharedSecret = const None(),
      this._issuer = const None(),
      HttpClient httpClient,
      this._audience = const None(),
      this._useAuthHeader = true,
      this._sendAddonAsSubject = false])
      : _httpClient = httpClient != null ? httpClient : new HttpClient();

  Future<HttpClientRequest> get(String path,
          {bool signRequest: false,
          Map queryParams: const {},
          String subject}) =>
      open(path, HttpMethod.GET,
          signRequest: signRequest, queryParams: queryParams, subject: subject);

  Future<HttpClientRequest> post(String path,
          {bool signRequest: false,
          Map queryParams: const {},
          String subject}) =>
      open(path, HttpMethod.POST,
          signRequest: signRequest, queryParams: queryParams, subject: subject);

  Future<HttpClientRequest> open(String path, HttpMethod method,
      {bool signRequest: false,
      Map queryParams: const {},
      String subject}) async {
    final request = await _httpClient.openUrl(
        method.literal,
        _createUri(path, method, signRequest && !_useAuthHeader, queryParams,
            subject));

    if (signRequest && _useAuthHeader) {
      final jwtToken = _createJwtToken(path, method, queryParams, subject);
      request.headers.add(HttpHeaders.AUTHORIZATION, 'JWT $jwtToken');
    }

    return request;
  }

  void addCredentials(String realm, HttpClientCredentials credentials) {
    _httpClient.addCredentials(baseUri, realm, credentials);
  }

  Uri _createUri(String path, HttpMethod method, bool signRequest,
      Map queryParams, String subject) {
    final Map params = signRequest ? new Map.from(queryParams) : queryParams;

    if (signRequest) {
      if (_sharedSecret is None || _issuer is None) {
        throw new ArgumentError("Cannot sign a request without a shared secret"
            " and issuer");
      }
      params['jwt'] = _createJwtToken(path, method, queryParams, subject);
    }

    Uri uri = Uri.parse(_url.join(_url.fromUri(baseUri), path));
    if (params.isNotEmpty) {
      uri = uri.scheme == 'http'
          ? new Uri.http(uri.authority, uri.path, params)
          : new Uri.https(uri.authority, uri.path, params);
    }

    _log.finest('sending ${method.literal} $uri');
    return uri;
  }

  String _createJwtToken(
      String path, HttpMethod method, Map queryParams, String subject) {
    final _subject =
        _sendAddonAsSubject ? _audience.getOrElse(() => null) : subject;
    final __audience =
        _sendAddonAsSubject ? null : _audience.getOrElse(() => null);

    return createProductHostToken(
        path, method.literal, queryParams, _sharedSecret.get(), _issuer.get(),
        subject: _subject, audience: __audience);
  }
}
