library stuff.client.rest;

import 'dart:async';
import 'dart:convert';

import 'package:http/http.dart' as http;
import 'package:logging/logging.dart';
import 'package:path/path.dart' as path;

final _p = path.url;

final Logger _log = new Logger('client.rest');

class ApiException implements Exception {}

class HttpException implements Exception, ApiException {
  final int statusCode;
  final String body;
  final String requestMethod;
  final Uri requestUrl;

  Map get bodyAsJson => JSON.decode(body);

  HttpException(
      this.statusCode, this.body, this.requestMethod, this.requestUrl);

  String toString() =>
      'Error HTTP response for ($requestMethod ${requestUrl.toString()}):'
      ' status: $statusCode; body: $body';
}

class UnexpectedRedirectException extends HttpException {
  final Uri redirectionUrl;

  UnexpectedRedirectException(int statusCode, String body, String requestMethod,
      Uri requestUrl, this.redirectionUrl)
      : super(statusCode, body, requestMethod, requestUrl);
}

class UnauthorisedException extends HttpException {
  UnauthorisedException(String body, String requestMethod, Uri requestUrl)
      : super(401, body, requestMethod, requestUrl);
}

class ForbiddenException extends HttpException {
  ForbiddenException(String body, String requestMethod, Uri requestUrl)
      : super(403, body, requestMethod, requestUrl);
}

class NotFoundException extends HttpException {
  NotFoundException(String body, String requestMethod, Uri requestUrl)
      : super(404, body, requestMethod, requestUrl);
}

class UnexpectedContentTypeException extends HttpException {
  final String contentType;

  UnexpectedContentTypeException(int statusCode, String body,
      String requestMethod, Uri requestUrl, this.contentType)
      : super(statusCode, body, requestMethod, requestUrl);
}

class RestResourceClient {
  final Uri _baseUri;
  final http.Client _client;
  final bool _isTest;

  @deprecated // hide this
  http.Client get client => _client;

  RestResourceClient._(this._baseUri, this._client, this._isTest);

  RestResourceClient(Uri baseUri, http.Client client)
      : this._(baseUri, client, false);

  RestResourceClient.test(Uri baseUri, http.Client client)
      : this._(baseUri, client, true);

  RestResourceClient childClient(String childPath) =>
      new RestResourceClient._(_absolute(childPath), _client, _isTest);

  Future<Map> getJson([String relativePath]) {
    return sendJson('GET', relativePath);
  }

  Future<Map> postJson(Map json, [String relativePath]) {
    return sendJson('POST', relativePath, json);
  }

  Future<Map> putJson(Map json, [String relativePath]) {
    return sendJson('PUT', relativePath, json);
  }

  Future<Map> sendJson(String method, String relativePath, [Map jsonBody]) {
    return send(method, relativePath, body: jsonBody);
  }

  Future<Map> send(String method, String relativePath,
      {Map body, Map headers: const {}, bool isForm: false}) {
    return doSend(method, relativePath,
        body: body, headers: headers, isForm: isForm);
  }

  Future<Map> doSend(String method, String relativePath,
      {Map body, Map headers: const {}, bool isForm: false}) {
    final uri = _absolute(relativePath);
    _log.finest('sending $method $uri');

    Map _headers = headers != null ? headers : {};

    final coreHeaders = {'accept': 'application/json'};

    if (!isForm) {
      coreHeaders['content-type'] = 'application/json';
    }

    final request = new http.Request(method, uri)
      ..headers.addAll(coreHeaders)
      ..headers.addAll(_headers)
      ..followRedirects = false;

    if (isForm) {
      request.bodyFields = body;
    } else {
      request.body = JSON.encode(body);
    }

    if (_isTest) {
      // set referer to avoid some origin issues
      request.headers['referer'] = _baseUri.toString();
    }

    return _client.send(request).then((response) {
      _log.finest('response headers ${response.headers}');
      final sc = response.statusCode;
      _log.finest('status code $sc');
      if (sc >= 200 && sc < 300) {
        return _jsonBody(response, method, uri);
      } else {
        return _handleErrorResponse(response, sc, method, uri);
      }
    });
  }

  Future<Map> sendFormBody(String method, String relativePath, Map body,
      {Map headers: const {}}) {
    return send(method, relativePath,
        body: body, isForm: true, headers: headers);
  }

  Future _handleErrorResponse(
      http.StreamedResponse response, int sc, String method, Uri uri) {
    _log.finest('_handleErrorResponse $sc: $method $uri');
    return _stringBody(response).then((body) {
      _log.finest('body: $body');
      if (sc >= 300 && sc < 400) {
        throw new UnexpectedRedirectException(
            sc, body, method, uri, Uri.parse(response.headers['location']));
      } else if (sc == 401) {
        throw new UnauthorisedException(body, method, uri);
      } else if (sc == 403) {
        throw new ForbiddenException(body, method, uri);
      } else if (sc == 404) {
        throw new NotFoundException(body, method, uri);
      } else {
        throw new HttpException(sc, body, method, uri);
      }
    });
  }

  Uri _absolute(String relativePath) {
    if (relativePath == null) {
      return _baseUri;
    }
    final adjustedPath = relativePath.startsWith(_p.separator)
        ? relativePath.replaceFirst('/', '')
        : relativePath;

    return _p.toUri(_p.join(_p.fromUri(_baseUri), adjustedPath));
  }
}

Future<Map> _jsonBody(http.StreamedResponse response, String method, Uri uri) {
  final contentType = response.headers['content-type'];

//  print('============= content-type: $contentType');

  if (contentType != null && !contentType.startsWith('application/json')) {
    return _stringBody(response).then((body) {
//      print('======== UnexpectedContentTypeException $body');
      throw new UnexpectedContentTypeException(response.statusCode, body,
          method, uri, response.headers['content-type']);
    });
  }
  return response.stream.toStringStream().transform(JSON.decoder).first;
}

Future<String> _stringBody(http.StreamedResponse response) {
  return response.stream.bytesToString();
}
